<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Basic Jembe application</title>
    <link rel="stylesheet" href="{{'{{'}} url_for('static', filename='css/{{project_name}}.css') {{'}}'}}">
    <link rel="stylesheet" href="https://fonts.googleapis.com/css?family=Roboto:300,300italic,700,700italic">
</head>
<body>
    <div>
        <h1>Welcome to "{{project_name}}"</h1>
        {{'{{'}}component("counter"){{'}}'}}
    </div>
    <script src="{{'{{'}} url_for('jembe.static', filename='js/jembe.js') {{'}}'}}"></script>
    <script defer>
    {# Add CSRF token to Jembe ajax requests  #}
    window.addEventListener('DOMContentLoaded', function(event){
        window.jembeClient.addXRequestHeaderGenerator(function () {
            return {'X-CSRFToken': window.jembeClient.getCookie("_csrf_token")};
        })
    })
    </script>
</body>
</html>