<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Jembe System Page</title>
</head>
<body style="box-sizing: border-box; margin:0px; background-color: #F5F5F5;">
    <div style="color: #4D7C0F; font-size: 85%; font-family: sans-serif; display:flex; flex-direction: column; width: 100vw; height: 100vh;">
        <div style="flex: 1 0 auto;"></div>
        <div style="flex-shrink:0;text-align: right; padding: 12px; ">
            Build with <a style="color: #4D7C0F;" href="https://jembe.io" target="_blank">jembe.io</a>.
        </div>
    <div>
   <script src="{{ url_for('jembe.static', filename='js/jembe.js') }}"></script>
</body>
</html>