from os import path

DEFAULT_JEMBE_MEDIA_FOLDER = path.join("..", "data", "media")
DEFAULT_TEMP_STORAGE_UPLOAD_FOLDER = "UPLOADS"
DEFAULT_SESSION_TEMP_STORAGE_ID = "jembe_temp_storage_id"
DEFAULT_SESSION_TEMP_STORAGE_SUBDIR = "WORKINPROGRESS"
DEFAULT_STORAGE_CACHE_FOLDER = "CACHE"

PUBLIC_STORAGE_NAME = "public"
PRIVATE_STORAGE_NAME = "private"
TEMP_STORAGE_NAME = "temp"